@@ -60,9 +60,71 @@
 #include "crypto-gnutls.h"
 #endif
 
-#ifdef WITH_SDP
-#include <sdp_inet.h>
-#endif
+#include "nbdclt.h"
+#include "nbdtab_parser.tab.h"
+
+CLIENT* cur_client;
+extern FILE *yyin, *yyout;
+bool found_config = false;
+bool parse_error = false;
+
+#define SET_PROP(str, member, rval) if(!strcmp(property, str)) { cur_client->member = (rval); return; }
+void nbdtab_set_property(char *property, char *val) {
+	if(found_config) return;
+	SET_PROP("port", port, val);
+	SET_PROP("certfile", cert, val);
+	SET_PROP("keyfile", key, val);
+	SET_PROP("cacertfile", cacert, val);
+	SET_PROP("tlshostname", tlshostn, val);
+	SET_PROP("bs", bs, strtol(val, NULL, 10));
+	SET_PROP("timeout", timeout, strtol(val, NULL, 10));
+	SET_PROP("conns", nconn, strtol(val, NULL, 10));
+	if(*property != '_') {
+		fprintf(stderr, "Warning: unknown option '%s' found in nbdtab file", property);
+	}
+}
+#undef SET_PROP
+
+#define SET_FLAG(str, member) if(!strcmp(property, str)) { cur_client->member = true; return; }
+void nbdtab_set_flag(char *property) {
+	if(found_config) return;
+	SET_FLAG("no_optgo", no_optgo);
+	SET_FLAG("persist", persist);
+	SET_FLAG("swap", swap);
+	SET_FLAG("unix", b_unix);
+	SET_FLAG("preinit", preinit);
+	SET_FLAG("tls", tls);
+	if(*property != '_') {
+		fprintf(stderr, "Warning: unknown option '%s' found in nbdtab file", property);
+	}
+}
+
+void nbdtab_commit_line(char *devn, char *hostn, char *exportname) {
+	if(!strncmp(devn, "/dev/", 5)) {
+		devn += 5;
+	}
+	if(!strcmp(cur_client->dev, devn)) {
+		found_config = true;
+		cur_client->hostn = hostn;
+		cur_client->name = exportname;
+	} else {
+		if(!found_config) {
+			char *tmp = cur_client->dev;
+			memset(cur_client, 0, sizeof(CLIENT));
+			cur_client->bs = 512;
+			cur_client->nconn = 1;
+			cur_client->dev = tmp;
+		}
+	}
+	return;
+}
+
+void yyerror(char *msg) {
+	parse_error = true;
+	fprintf(stderr, "parse error parsing " SYSCONFDIR "/nbdtab: %s", msg);
+}
+
+#undef SET_FLAG
 
 #define NBDC_DO_LIST 1
 
@@ -120,7 +182,8 @@ static void netlink_configure(int index, int *sockfds, int num_connects,
 	NLA_PUT_U64(msg, NBD_ATTR_SIZE_BYTES, size64);
 	NLA_PUT_U64(msg, NBD_ATTR_BLOCK_SIZE_BYTES, blocksize);
 	NLA_PUT_U64(msg, NBD_ATTR_SERVER_FLAGS, flags);
-	NLA_PUT_U64(msg, NBD_ATTR_TIMEOUT, timeout);
+	if (timeout)
+		NLA_PUT_U64(msg, NBD_ATTR_TIMEOUT, timeout);
 
 	sock_attr = nla_nest_start(msg, NBD_ATTR_SOCKETS);
 	if (!sock_attr)
@@ -215,7 +278,7 @@ int check_conn(char* devname, int do_print) {
 	return 0;
 }
 
-int opennet(char *name, char* portstr, int sdp) {
+int opennet(char *name, char* portstr) {
 	int sock;
 	struct addrinfo hints;
 	struct addrinfo *ai = NULL;
@@ -236,17 +299,6 @@ int opennet(char *name, char* portstr, int sdp) {
 		return -1;
 	}
 
-	if(sdp) {
-#ifdef WITH_SDP
-		if (ai->ai_family == AF_INET)
-			ai->ai_family = AF_INET_SDP;
-		else (ai->ai_family == AF_INET6)
-			ai->ai_family = AF_INET6_SDP;
-#else
-		err("Can't do SDP: I was not compiled with SDP support!");
-#endif
-	}
-
 	for(rp = ai; rp != NULL; rp = rp->ai_next) {
 		sock = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
 
@@ -461,7 +513,20 @@ void parse_sizes(char *buf, uint64_t *size, uint16_t *flags) {
 	printf("\n");
 }
 
-void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t needed_flags, uint32_t client_flags, uint32_t do_opts, char *certfile, char *keyfile, char *cacertfile, char *tlshostname, bool tls) {
+void send_opt_exportname(int sock, u64 *rsize64, uint16_t *flags, bool can_opt_go, char* name, uint16_t global_flags) {
+	send_request(sock, NBD_OPT_EXPORT_NAME, -1, name);
+	char b[sizeof(*flags) + sizeof(*rsize64)];
+	if(readit(sock, b, sizeof(b)) < 0 && can_opt_go) {
+		err("E: server does not support NBD_OPT_GO and dropped connection after sending NBD_OPT_EXPORT_NAME. Try -g.");
+	}
+	parse_sizes(b, rsize64, flags);
+	if(!(global_flags & NBD_FLAG_NO_ZEROES)) {
+		char buf[125];
+		readit(sock, buf, 124);
+	}
+}
+
+void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t needed_flags, uint32_t client_flags, uint32_t do_opts, char *certfile, char *keyfile, char *cacertfile, char *tlshostname, bool tls, bool can_opt_go) {
 	u64 magic;
 	uint16_t tmp;
 	uint16_t global_flags;
@@ -594,25 +659,24 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 		exit(EXIT_SUCCESS);
 	}
 
-	send_info_request(sock, NBD_OPT_GO, 0, NULL, name);
-
 	struct reply *rep = NULL;
 	
+	if(!can_opt_go) {
+		send_opt_exportname(sock, rsize64, flags, can_opt_go, name, global_flags);
+		return;
+	}
+
+	send_info_request(sock, NBD_OPT_GO, 0, NULL, name);
+
 	do {
 		if(rep != NULL) free(rep);
 		rep = read_reply(sock);
-		if(rep->reply_type & NBD_REP_FLAG_ERROR) {
+		if(rep && (rep->reply_type & NBD_REP_FLAG_ERROR)) {
 			switch(rep->reply_type) {
 				case NBD_REP_ERR_UNSUP:
 					/* server doesn't support NBD_OPT_GO or NBD_OPT_INFO,
 					 * fall back to NBD_OPT_EXPORT_NAME */
-					send_request(sock, NBD_OPT_EXPORT_NAME, -1, name);
-					char b[sizeof(*flags) + sizeof(*rsize64)];
-					readit(sock, b, sizeof(b));
-					parse_sizes(b, rsize64, flags);
-					if(!(global_flags & NBD_FLAG_NO_ZEROES)) {
-						readit(sock, buf, 124);
-					}
+					send_opt_exportname(sock, rsize64, flags, can_opt_go, name, global_flags);
 					free(rep);
 					return;
 				case NBD_REP_ERR_POLICY:
@@ -623,6 +687,7 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 					} else {
 						err("Connection not allowed by server policy.");
 					}
+					free(rep);
 					exit(EXIT_FAILURE);
 				default:
 					if(rep->datasize > 0) {
@@ -653,141 +718,57 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 			case NBD_REP_ACK:
 				break;
 			default:
-				err_nonfatal("Unknown reply to NBD_OPT_GO received");
+				err_nonfatal("Unknown reply to NBD_OPT_GO received, ignoring");
 		}
 	} while(rep->reply_type != NBD_REP_ACK);
 	free(rep);
 }
 
-bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** hostn_ptr, int* bs, int* timeout, int* persist, int* swap, int* sdp, int* b_unix, char**port, int* num_conns, char **certfile, char **keyfile, char **cacertfile, char **tlshostname) {
-	int fd = open(SYSCONFDIR "/nbdtab", O_RDONLY);
+bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** hostn_ptr, int* bs, int* timeout, int* persist, int* swap, int* b_unix, char**port, int* num_conns, char **certfile, char **keyfile, char **cacertfile, char **tlshostname, bool *can_opt_go) {
 	bool retval = false;
-	if(fd < 0) {
+	cur_client = calloc(sizeof(CLIENT), 1);
+	cur_client->bs = 512;
+	cur_client->nconn = 1;
+	yyin = fopen(SYSCONFDIR "/nbdtab", "r");
+	yyout = fopen("/dev/null", "w");
+	
+	if(!strncmp(cfgname, "/dev/", 5)) {
+		cfgname += 5;
+	}
+	cur_client->dev = cfgname;
+	if(yyin == NULL) {
 		fprintf(stderr, "while opening %s: ", SYSCONFDIR "/nbdtab");
 		perror("could not open config file");
 		goto out;
 	}
-	off_t size = lseek(fd, 0, SEEK_END);
-	lseek(fd, 0, SEEK_SET);
-	void *data = NULL;
-	char *fsep = "\n\t# ";
-	char *lsep = "\n#";
+	yyparse();
 
-	if(size < 0) {
-		perror("E: mmap'ing nbdtab");
-		exit(EXIT_FAILURE);
-	}
-
-	data = mmap(NULL, (size_t)size, PROT_READ, MAP_SHARED, fd, 0);
-	if(!strncmp(cfgname, "/dev/", 5)) {
-		cfgname += 5;
-	}
-	char *loc = strstr((const char*)data, cfgname);
-	if(!loc) {
+	if(!found_config || parse_error) {
 		goto out;
 	}
-	size_t l = strlen(cfgname) + 6;
-	*dev_ptr = malloc(l);
-	snprintf(*dev_ptr, l, "/dev/%s", cfgname);
+	*name_ptr = cur_client->name;
+	*dev_ptr = calloc(strlen(cur_client->dev) + 6, 1);
+	snprintf(*dev_ptr, strlen(cur_client->dev) + 6, "/dev/%s", cur_client->dev);
+	*hostn_ptr = cur_client->hostn;
+	*bs = cur_client->bs;
+	*timeout = cur_client->timeout;
+	*persist = cur_client->persist ? 1 : 0;
+	*swap = cur_client->swap ? 1 : 0;
+	*b_unix = cur_client->b_unix ? 1 : 0;
+	*num_conns = cur_client->nconn;
+	*certfile = cur_client->cert;
+	*keyfile = cur_client->key;
+	*cacertfile = cur_client->cacert;
+	*tlshostname = cur_client->tlshostn;
+	*can_opt_go = !(cur_client->no_optgo);
 
-	size_t line_len, field_len, ws_len;
-#define CHECK_LEN field_len = strcspn(loc, fsep); ws_len = strspn(loc+field_len, fsep); if(field_len > line_len || line_len <= 0) { goto out; }
-#define MOVE_NEXT line_len -= field_len + ws_len; loc += field_len + ws_len
-	// find length of line
-	line_len = strcspn(loc, lsep);
-	// first field is the device node name, which we already know, so skip it
-	CHECK_LEN;
-	MOVE_NEXT;
-	// next field is the hostname
-	CHECK_LEN;
-	*hostn_ptr = strndup(loc, field_len);
-	MOVE_NEXT;
-	// third field is the export name
-	CHECK_LEN;
-	*name_ptr = strndup(loc, field_len);
-	if(ws_len + field_len > line_len) {
-		// optional last field is not there, so return success
-		retval = true;
-		goto out;
-	}
-	MOVE_NEXT;
-	CHECK_LEN;
-#undef CHECK_LEN
-#undef MOVE_NEXT
-	// fourth field is the options field, a comma-separated field of options
-	do {
-		if(!strncmp(loc, "conns=", 6)) {
-			*num_conns = (int)strtol(loc+6, &loc, 0);
-			goto next;
-		}
-		if(!strncmp(loc, "bs=", 3)) {
-			*bs = (int)strtol(loc+3, &loc, 0);
-			goto next;
-		}
-		if(!strncmp(loc, "timeout=", 8)) {
-			*timeout = (int)strtol(loc+8, &loc, 0);
-			goto next;
-		}
-		if(!strncmp(loc, "port=", 5)) {
-			*port = strndup(loc+5, strcspn(loc+5, ","));
-			goto next;
-		}
-		if(!strncmp(loc, "persist", 7)) {
-			loc += 7;
-			*persist = 1;
-			goto next;
-		}
-		if(!strncmp(loc, "swap", 4)) {
-			*swap = 1;
-			loc += 4;
-			goto next;
-		}
-		if(!strncmp(loc, "sdp", 3)) {
-			*sdp = 1;
-			loc += 3;
-			goto next;
-		}
-		if(!strncmp(loc, "unix", 4)) {
-			*b_unix = 1;
-			loc += 4;
-			goto next;
-		}
-		if(!strncmp(loc, "certfile=", 9)) {
-			*certfile = strndup(loc+9, strcspn(loc+9, ","));
-			goto next;
-		}
-		if(!strncmp(loc, "keyfile=", 8)) {
-			*keyfile = strndup(loc+8, strcspn(loc+8, ","));
-			goto next;
-		}
-		if(!strncmp(loc, "cacertfile=", 11)) {
-			*cacertfile = strndup(loc+11, strcspn(loc+11, ","));
-			goto next;
-		}
-		if(!strncmp(loc, "tlshostname=", 9)) {
-			*tlshostname = strndup(loc+9, strcspn(loc+9, ","));
-			goto next;
-		}
-		// skip unknown options, with a warning unless they start with a '_'
-		l = strcspn(loc, ",");
-		if(*loc != '_') {
-			char* s = strndup(loc, l);
-			fprintf(stderr, "Warning: unknown option '%s' found in nbdtab file", s);
-			free(s);
-		}
-		loc += l;
-next:
-		if(*loc == ',') {
-			loc++;
-		}
-	} while(strcspn(loc, lsep) > 0);
 	retval = true;
 out:
-	if(data != NULL) {
-		munmap(data, size);
+	if(yyin != NULL) {
+		fclose(yyin);
 	}
-	if(fd >= 0) {
-		close(fd);
+	if(yyout != NULL) {
+		fclose(yyout);
 	}
 	return retval;
 }
@@ -878,9 +859,9 @@ void usage(char* errmsg, ...) {
 		fprintf(stderr, "%s version %s\n", PROG_NAME, PACKAGE_VERSION);
 	}
 #if HAVE_NETLINK
-	fprintf(stderr, "Usage: nbd-client -name|-N name host [port] nbd_device\n\t[-block-size|-b block size] [-timeout|-t timeout] [-swap|-s] [-sdp|-S]\n\t[-persist|-p] [-nofork|-n] [-systemd-mark|-m] [-nonetlink|-L]\n");
+	fprintf(stderr, "Usage: nbd-client -name|-N name host [port] nbd_device\n\t[-block-size|-b block size] [-timeout|-t timeout] [-swap|-s]\n\t[-persist|-p] [-nofork|-n] [-systemd-mark|-m] [-nonetlink|-L]\n\t[-readonly|-R] [-size|-B bytes] [-preinit|-P]\n");
 #else
-	fprintf(stderr, "Usage: nbd-client -name|-N name host [port] nbd_device\n\t[-block-size|-b block size] [-timeout|-t timeout] [-swap|-s] [-sdp|-S]\n\t[-persist|-p] [-nofork|-n] [-systemd-mark|-m]\n");
+	fprintf(stderr, "Usage: nbd-client -name|-N name host [port] nbd_device\n\t[-block-size|-b block size] [-timeout|-t timeout] [-swap|-s]\n\t[-persist|-p] [-nofork|-n] [-systemd-mark|-m]\n\t[-readonly|-R] [-size|-B bytes] [-preinit|-P]\n");
 #endif
 	fprintf(stderr, "Or   : nbd-client -u (with same arguments as above)\n");
 	fprintf(stderr, "Or   : nbd-client nbdX\n");
@@ -892,7 +873,7 @@ void usage(char* errmsg, ...) {
 #if HAVE_GNUTLS && !defined(NOTLS)
 	fprintf(stderr, "All commands that connect to a host also take:\n\t[-F|-certfile certfile] [-K|-keyfile keyfile]\n\t[-A|-cacertfile cacertfile] [-H|-tlshostname hostname] [-x|-enable-tls]\n");
 #endif
-	fprintf(stderr, "Default value for blocksize is 1024 (recommended for ethernet)\n");
+	fprintf(stderr, "Default value for blocksize is 512\n");
 	fprintf(stderr, "Allowed values for blocksize are 512,1024,2048,4096\n"); /* will be checked in kernel :) */
 	fprintf(stderr, "Note, that kernel 2.4.2 and older ones do not work correctly with\n");
 	fprintf(stderr, "blocksizes other than 1024 without patches\n");
@@ -912,28 +893,31 @@ void disconnect(char* device) {
 	if (ioctl(nbd, NBD_CLEAR_SOCK)<0)
 		err("Ioctl failed: %m\n");
 	printf("done\n");
+	close(nbd);
 }
 
 #if HAVE_NETLINK
-static const char *short_opts = "-A:b:c:C:d:H:hK:LlnN:pSst:uVx";
+static const char *short_opts = "-A:B:b:c:C:d:gH:hK:LlnN:PpRSst:uVx";
 #else
-static const char *short_opts = "-A:b:c:C:d:H:hK:lnN:pSst:uVx";
+static const char *short_opts = "-A:B:b:c:C:d:gH:hK:lnN:PpRSst:uVx";
 #endif
 
 int main(int argc, char *argv[]) {
 	char* port=NBD_DEFAULT_PORT;
 	int sock, nbd;
-	int blocksize=1024;
+	int blocksize=512;
 	char *hostname=NULL;
 	char *nbddev=NULL;
 	int swap=0;
 	int cont=0;
 	int timeout=0;
-	int sdp=0;
 	int G_GNUC_UNUSED nofork=0; // if -dNOFORK
 	pid_t main_pid;
-	u64 size64;
+	u64 size64 = 0;
+	u64 force_size64 = 0;
 	uint16_t flags = 0;
+	bool force_read_only = false;
+	bool preinit = false;
 	int c;
 	int nonspecial=0;
 	int b_unix=0;
@@ -953,32 +937,36 @@ int main(int argc, char *argv[]) {
 	int need_disconnect = 0;
 	int *sockfds;
 	struct option long_options[] = {
+		{ "cacertfile", required_argument, NULL, 'A' },
 		{ "block-size", required_argument, NULL, 'b' },
+		{ "size", required_argument, NULL, 'B' },
 		{ "check", required_argument, NULL, 'c' },
 		{ "connections", required_argument, NULL, 'C'},
 		{ "disconnect", required_argument, NULL, 'd' },
+		{ "certfile", required_argument, NULL, 'F' },
+		{ "no-optgo", no_argument, NULL, 'g' },
 		{ "help", no_argument, NULL, 'h' },
+		{ "tlshostname", required_argument, NULL, 'H' },
+		{ "keyfile", required_argument, NULL, 'K' },
 		{ "list", no_argument, NULL, 'l' },
-		{ "name", required_argument, NULL, 'N' },
 #if HAVE_NETLINK
 		{ "nonetlink", no_argument, NULL, 'L' },
 #endif
-		{ "nofork", no_argument, NULL, 'n' },
-		{ "persist", no_argument, NULL, 'p' },
-		{ "sdp", no_argument, NULL, 'S' },
-		{ "swap", no_argument, NULL, 's' },
 		{ "systemd-mark", no_argument, NULL, 'm' },
+		{ "nofork", no_argument, NULL, 'n' },
+		{ "name", required_argument, NULL, 'N' },
+		{ "persist", no_argument, NULL, 'p' },
+		{ "preinit", no_argument, NULL, 'P' },
+		{ "readonly", no_argument, NULL, 'R' },
+		{ "swap", no_argument, NULL, 's' },
 		{ "timeout", required_argument, NULL, 't' },
 		{ "unix", no_argument, NULL, 'u' },
-		{ "certfile", required_argument, NULL, 'F' },
-		{ "keyfile", required_argument, NULL, 'K' },
-		{ "cacertfile", required_argument, NULL, 'A' },
-		{ "tlshostname", required_argument, NULL, 'H' },
-		{ "enable-tls", no_argument, NULL, 'x' },
 		{ "version", no_argument, NULL, 'V' },
+		{ "enable-tls", no_argument, NULL, 'x' },
 		{ 0, 0, 0, 0 }, 
 	};
 	int i;
+	bool can_opt_go = true;
 
 	logging(MY_NAME);
 
@@ -1032,6 +1020,17 @@ int main(int argc, char *argv[]) {
 		case 'b':
 		      blocksize:
 			blocksize=(int)strtol(optarg, NULL, 0);
+			if(blocksize == 0 || (blocksize % 512) != 0) {
+				fprintf(stderr, "E: blocksize is not a multiple of 512! This is not allowed\n");
+				exit(EXIT_FAILURE);
+			}
+			break;
+		case 'B':
+			force_size64=(u64)strtoull(optarg, NULL, 0);
+			if(force_size64 == 0) {
+				fprintf(stderr, "E: Invalid size\n");
+				exit(EXIT_FAILURE);
+			}
 			break;
 		case 'c':
 			return check_conn(optarg, 1);
@@ -1042,6 +1041,9 @@ int main(int argc, char *argv[]) {
 			need_disconnect = 1;
 			nbddev = strdup(optarg);
 			break;
+		case 'g':
+			can_opt_go = false;
+			break;
 		case 'h':
 			usage(NULL);
 			exit(EXIT_SUCCESS);
@@ -1067,12 +1069,15 @@ int main(int argc, char *argv[]) {
 		case 'p':
 			cont=1;
 			break;
+		case 'P':
+			preinit = true;
+			break;
+		case 'R':
+			force_read_only = true;
+			break;
 		case 's':
 			swap=1;
 			break;
-		case 'S':
-			sdp=1;
-			break;
 		case 't':
 		      timeout:
 			timeout=strtol(optarg, NULL, 0);
@@ -1127,7 +1132,7 @@ int main(int argc, char *argv[]) {
 	if(hostname) {
 		if((!name || !nbddev) && !(opts & NBDC_DO_LIST)) {
 			if(!strncmp(hostname, "nbd", 3) || !strncmp(hostname, "/dev/nbd", 8)) {
-				if(!get_from_config(hostname, &name, &nbddev, &hostname, &blocksize, &timeout, &cont, &swap, &sdp, &b_unix, &port, &num_connections, &certfile, &keyfile, &cacertfile, &hostname)) {
+				if(!get_from_config(hostname, &name, &nbddev, &hostname, &blocksize, &timeout, &cont, &swap, &b_unix, &port, &num_connections, &certfile, &keyfile, &cacertfile, &tlshostname, &can_opt_go)) {
 					usage("no valid configuration for specified device found", hostname);
 					exit(EXIT_FAILURE);
 				}
@@ -1148,12 +1153,28 @@ int main(int argc, char *argv[]) {
 		tls = true;
 	}
 
-        if (!tlshostname && hostname)
+	if (preinit) {
+		if (tls) {
+			fprintf(stderr, "E: preinit connection cannot be used with TLS\n");
+			exit(EXIT_FAILURE);
+		}
+		if (!force_size64) {
+			fprintf(stderr, "E: preinit connection requires specifying size\n");
+			exit(EXIT_FAILURE);
+		}
+	}
+
+        if (!tlshostname && hostname && !b_unix)
                 tlshostname = strdup(hostname);
 
 	if (netlink)
 		nofork = 1;
 
+	if((force_size64 % blocksize) != 0) {
+		fprintf(stderr, "E: size (%" PRIu64 " bytes) is not a multiple of blocksize (%d)!\n", force_size64, blocksize);
+		exit(EXIT_FAILURE);
+	}
+
 	if(strlen(name)==0 && !(opts & NBDC_DO_LIST)) {
 		printf("Warning: the oldstyle protocol is no longer supported.\nThis method now uses the newstyle protocol with a default export\n");
 	}
@@ -1174,11 +1195,16 @@ int main(int argc, char *argv[]) {
 		if (b_unix)
 			sock = openunix(hostname);
 		else
-			sock = opennet(hostname, port, sdp);
+			sock = opennet(hostname, port);
 		if (sock < 0)
 			exit(EXIT_FAILURE);
 
-		negotiate(&sock, &size64, &flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls);
+		if (!preinit)
+			negotiate(&sock, &size64, &flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls, can_opt_go);
+		if (force_read_only)
+			flags |= NBD_FLAG_READ_ONLY;
+		if (force_size64)
+			size64 = force_size64;
 		if (netlink) {
 			sockfds[i] = sock;
 			continue;
@@ -1268,7 +1294,7 @@ int main(int argc, char *argv[]) {
 
 		if (ioctl(nbd, NBD_DO_IT) < 0) {
 			int error = errno;
-			fprintf(stderr, "nbd,%d: Kernel call returned: %d", main_pid, error);
+			fprintf(stderr, "nbd,%d: Kernel call returned: %s\n", main_pid, strerror(errno));
 			if(error==EBADR) {
 				/* The user probably did 'nbd-client -d' on us.
 				 * quit */
@@ -1284,7 +1310,7 @@ int main(int argc, char *argv[]) {
 						if (b_unix)
 							sock = openunix(hostname);
 						else
-							sock = opennet(hostname, port, sdp);
+							sock = opennet(hostname, port);
 						if (sock >= 0)
 							break;
 						sleep (1);
@@ -1292,7 +1318,7 @@ int main(int argc, char *argv[]) {
 					nbd = open(nbddev, O_RDWR);
 					if (nbd < 0)
 						err("Cannot open NBD: %m");
-					negotiate(&sock, &new_size, &new_flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls);
+					negotiate(&sock, &new_size, &new_flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls, can_opt_go);
 					if (size64 != new_size) {
 						err("Size of the device changed. Bye");
 					}
@@ -1307,7 +1333,7 @@ int main(int argc, char *argv[]) {
 			/* We're on 2.4. It's not clearly defined what exactly
 			 * happened at this point. Probably best to quit, now
 			 */
-			fprintf(stderr, "Kernel call returned.");
+			fprintf(stderr, "Kernel call returned.\n");
 			cont=0;
 		}
 	} while(cont);
